#ifndef __DEBUG_H_INCLUDED__
#define __DEBUG_H_INCLUDED__

// debugger

#include "OurTypes.h"
#include "pilotcpu\pilotcpu.h"

const int kcpszArgsMax = 50;
const int kcchSymbolMax = 50;

typedef ulong LocalID;

typedef DWORD VirtualAddress;   // va
const VirtualAddress kvaInvalid = (VirtualAddress)-1;
const VirtualAddress kvaNextAppStartup = 0x10c0c5b8;

// NOTE: The VariableList class is probably temporary. If any other decent
// collection classes are used then they can replace it.

struct Variable {
    struct Variable *pvarNext;
    char *pszName;
    char *pszValue;
};

class VariableList {
public:
    VariableList();
    ~VariableList();
    BOOL Set(const char *pszName, const char *pszValue);
    char *Lookup(const char *pszName);
    Variable *GetNext(Position *ppos);

private:
    Variable *Lookup(const char *pszName, Variable ***pppvar);

private:
    Variable *m_pvarHead;
};


enum DbgErr {
    kerrNone, kerrSyntax, kerrDivideByZero, kerrUndefinedSymbol,
    kerrNumberTooBig, kerrAsciiTooBig
};
const int kclStackMax = 5;

//
// Modules & symbols
//

struct Symbol {         // sym
    VirtualAddress va;
	VirtualAddress va_RTS;
    char *pszName;
};

class Module {          // mod
    friend class Debugger;

private:
    static Module *m_pmodHead;
    class Module *m_pmodNext;
    char m_pszName[ 32 ];
	union {
		DWORD dw;
		CHAR ach[4];
	} m_Type;
	union {
		DWORD dw;
		CHAR ach[4];
	} m_Creator;

    LocalID m_lidDb;
    long m_csym;              // count of symbols
    Symbol *m_asym;           // array of symbols
    VirtualAddress m_vaCode;
    long m_cbCode;

public:
	static Module * Next( Module *pmod );
    Module(LocalID lidDb, char *pszName, DWORD Type, DWORD Creator);
    ~Module();
    Symbol *FindSymbol(const char *pszName);
    Symbol *FindSymbol(VirtualAddress va);
    BOOL FindNearestSymbols(VirtualAddress va,
            char **ppszBefore, long *pcbBOffset, char **ppszAfter, 
            long *pcbAOffset);
    BOOL GetSymbols(VirtualAddress vaCode, long cbCode);
	BOOL DumpSymbols(Debugger *pdbg);

	const char * const GetQualifiedName(void);

	static Module *FindModuleByTypeCreator( DWORD Type, DWORD Creator );
	static DWORD TypeFromSymbol( const char *pszSymbol );
	static DWORD CreatorFromSymbol( const char *pszSymbol );
	static Module *FindModuleByAddr(VirtualAddress va);

	long GetCodeSize( void ) { return( m_cbCode ); };
	VirtualAddress GetStartAddress( void ) { return( m_vaCode ); };
	const char * GetName( void ) { return( m_pszName ); };
};


// UNDONE: Who calls this? Why? etc

class Debugger {
public:
	void BreakCpu( void );
	void Pause( DWORD dwMilliseconds );
	void Resume( DWORD dwMilliseconds );

	Debugger(BOOL (__cdecl * pfnOutput)(char *pszOutput));

    void Run(VirtualAddress vaStop);
    void Step();

    BOOL Execute(const char *pszCmd);
    BOOL printf(const char *pszFormat, ...);
    void DumpCpuState(VirtualAddress *pvaNextPC = NULL);
    void Disassemble(VirtualAddress vaDis, int cLines,
            VirtualAddress *pvaNextDis = NULL);

    BOOL SetVariable(const char *pszName, const char *pszValue);
    BOOL SetVariable(const char *pszName, ulong ulValue);
    BOOL GetVariable(const char *pszName, char *pszValue, int cchValueMax);
    BOOL GetVariable(const char *pszName, ulong *pulValue);
    const char *GetVariablePtr(const char *pszName);
    Variable *GetNextVariable(Position *ppos);

    Module *AddModule(LocalID lidDb);
	Module *AddModule( Module *pmod );

    BOOL RemoveModule(LocalID lidDb);
    Module *FindModule(LocalID lidDb);

    const char *Evaluate(const char *pszExpr, long *plValue, DbgErr *perr);
    DbgErr EvalVariable(const char *pszName, long *plValue);
    const char *LookupSymbol(VirtualAddress va);
    VirtualAddress LookupAddress(const char *pszSymbol);
    BOOL FindNearestSymbols(VirtualAddress va, char **ppszBefore = NULL, 
            long *pcbBOffset = NULL, char **ppszAfter = NULL, 
            long *pcbAOffset = NULL);

    int GetDefaultBase();
    void SetDefaultBase(int nBase);

private:
    const char *EvalNumber(const char *pszExpr, long *plResult, DbgErr *perr);

private:
	bool m_bBreakRequested;
	HANDLE evtDebuggerPause;
	HANDLE evtDebuggerResume;
    BOOL (__cdecl * m_pfnOutput)(char *pszOutput);
    VariableList m_vl;
};

extern ulong GetPC();
extern ulong *GetRegPtr(CPU_regs &regs, const char *pszName);
extern int GetInstructionLength(Debugger *pdbg, VirtualAddress va);
extern BOOL DebugCommandHandler(Debugger *pdbg, char *apszArgs[], int cpszArgs);

#define kszNextDisasm "_NextDisasm"
#define kszNextDump "_NextDump"
#define kszNumLines "_NumLines"
#define kszBreakCmd "_BreakCmd"
#define kszAppStartup "_AppStartup"

//
// Breakpoint stuff
//

const ushort kopcBreak = 0x4AFC;
const ushort kopcRts = 0x4e75;
const ushort kopcLink = 0x4e56;

const long kcbpUserMax = 20;
const long kcbpMax = 40;

typedef long Bpid;      // bpid
const Bpid kbpidGo = kcbpUserMax;
const Bpid kbpidSysAppLaunch = kcbpUserMax + 1;
const Bpid kbpidSysAppLaunchRts = kcbpUserMax + 2;
const Bpid kbpidSysAppLaunchCallApp = kcbpUserMax + 3;
const Bpid kbpidNextAppStartup = kcbpUserMax + 5;
const Bpid kbpidAppEntrypoint = kcbpUserMax + 6;
const Bpid kbpidAll = -1;
const Bpid kbpidUnspecified = -2;

typedef BOOL (*Pfnbph)(Debugger *pdbg, Bpid bpid);

struct Breakpoint {     // bp
    VirtualAddress va;
    ushort ff;
    ushort usInstr;
    Pfnbph pfnbph;
};

const ushort kfBpInUse = 0x0001;
const ushort kfBpEnabled = 0x0002;

BOOL SetBreakpoint(VirtualAddress va, Bpid bpid = kbpidUnspecified, 
        Pfnbph pfnbph = NULL);
BOOL ClearBreakpoint(Bpid bpid);
BOOL EnableBreakpoint(Bpid bpid, BOOL fEnable);
void InstallBreakpoints();
void RemoveBreakpoints();
BOOL DoBreakpoints(Debugger *pdbg);
void ShowBreakpoints(Debugger *pdbg);

//
// Some handy helpers
//

inline VirtualAddress GetHandlePtr(ulong h)
{
    return CPU_getlong(h & 0x7fffffff);
}

inline long GetHandleSize(ulong h)
{
    VirtualAddress va = GetHandlePtr(h);
    byte flags = CPU_getbyte(va - 6 + 3);
    return (CPU_getword(va - 6) - (flags & 0xf)) - 6;
}

const DWORD ROM_TYPE = '\0mor';
const DWORD ROM_CREATOR = '\0mor';

extern void StartDebugger(void); // this might come from any debugger
extern void StopDebugger(void); // this might come from any debugger
extern void ShowDebugger( bool bStopCpu = true );

#endif // __DEBUG_H_INCLUDED__
